/**
 * @brief Provides screen I/O functionality.
 *
 * High-level screen I/O functions and types.
 */

#pragma once

#include <atomic>
#include <cstdint>
#include <functional>
#include <iostream>
#include <mutex>
#include <streambuf>

namespace neovision {

// Classes/enums/structs/... ---------------------------------------------------

/**
 * @brief Standard 16 color definitions.
 */
enum class StandardColor {
    BLACK,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    MAGENTA,
    CYAN,
    WHITE,
    BRIGHT_BLACK,
    BRIGHT_RED,
    BRIGHT_GREEN,
    BRIGHT_YELLOW,
    BRIGHT_BLUE,
    BRIGHT_MAGENTA,
    BRIGHT_CYAN,
    BRIGHT_WHITE
};

/**
 * @brief 2d vector holding a x and y values.
 *
 * Typically holds the columns (X) and rows (Y) of a screen position or size.
 */
class Vector2D {

    std::atomic<std::uint32_t> m_x{0};
    std::atomic<std::uint32_t> m_y{0};

public:

    /**
     * @brief Default constructor.
     */
    Vector2D();

    /**
     * @brief Constructor with unsigned initial values.
     */
    Vector2D(std::uint32_t x, std::uint32_t y);

    /**
     * @brief Copy constructor.
     *
     * @param other Vector to be copied.
     */
    Vector2D(const Vector2D& other);

    /**
     * @brief Get value as string.
     * @return Returns the x and y coordinates in the form of "x,y"
     */
    std::string Str() const;

    /**
     * @brief Get X.
     * @return Returns the vector x coordinate.
     */
    std::uint32_t X() const;

    /**
     * @brief Set X.
     * @param n New x value.
     */
    void X(std::uint32_t n);

    /**
     * @brief Get Y.
     * @return Returns the vector y coordinate.
     */
    std::uint32_t Y() const;

    /**
     * @brief Set Y.
     * @param n New y value.
     */
    void Y(std::uint32_t n);

    Vector2D operator+(const Vector2D& other) const;
    Vector2D operator+=(const Vector2D& other);
    Vector2D operator-(const Vector2D& other) const;
    Vector2D operator-=(const Vector2D& other);
    void operator=(const Vector2D& other);
    bool operator==(const Vector2D& other) const;
    bool operator!=(const Vector2D& other) const;

};

/**
 * @brief Callback signature for IO read operations.
 *
 * The function should return the read bytes, and takes as parameter the number
 * of bytes that should be read.
 *
 */
typedef std::function<std::wstring(size_t)> IoReadFunction;

/**
 * @brief Callback signature for IO write operations.
 *
 * The function takes as parameter the data that should be written to the
 * terminal.
 */
typedef std::function<void(const std::wstring&)> IoWriteFunction;

/**
 * @brief Input/Output.
 *
 * Internal singleton class to hold parameters and functions for thread-safe 
 * text input and output.
 *
 * You typically shouldn't have to interact with this unless you want to change
 * how input/output to the terminal works. For instance, if you were to want to
 * output to stderr, you could do the following:
 *
 * @code{.cpp}
 * IO::get().SetOutputFunction([](std::wstring& s){ std::cerr << s; });
 * @endcode
 *
 */
class IO {

    std::mutex m_inputMutex;
    std::mutex m_outputMutex;
    IoReadFunction m_readFunction;
    IoWriteFunction m_writeFunction;

public:

    /**
     * @brief Default constructor.
     */
    IO();

    /**
     * @brief Destructor.
     */
    ~IO();

    /**
     * @brief Copy constructor (deleted).
     */
    IO(IO const&) = delete;

    /**
     * @brief Assignment operator (deleted).
     */
    void operator=(IO const&) = delete;

    /**
     * @brief Get singleton instance.
     *
     * @return Returns an instance of OutputParameters.
     */
    static IO& Get();

    /**
     * Lock input stream.
     *
     * Locks the current input stream. No input can happen until 
     * UnlockInputStream() is called.
     */
    void LockInputStream();

    /**
     * Lock output stream.
     *
     * Locks the current output stream. No input can happen until 
     * UnlockOutputStream() is called.
     */
    void LockOutputStream();

    /**
     * @brief Reads from the current input stream.
     *
     * Reads given number of bytes off of the input stream.
     *
     * @param bytes Number of bytes to read off of the stream.
     * @return Returns the data read from the stream.
     */
    std::wstring Read(std::size_t bytes);

    /**
     * @brief Reads from the current input stream.
     *
     * Reads bytes until given character is encountered and returns the data
     * read up until (and including) the given character.
     *
     * @param c Character to read up to.
     * @return Returns the data read from the stream.
     */
    std::wstring ReadUntil(wchar_t c);

    /**
     * @brief Sets the current input function.
     *
     * By default this class reads from stdin, you can override this by setting
     * this function to a callback that reads data from the terminal.
     *
     * @param f The new input function to be used for text input.
     */
    void SetInputFunction(IoReadFunction f);

    /**
     * @brief Sets the current output function.
     *
     * By default this class writes to stdout, you can override this by setting
     * this function to a callback that writes data to the terminal.
     *
     * @param f The new output function to be used for text output.
     */
    void SetOutputFunction(IoWriteFunction f);

    /**
     * Unlock input stream.
     *
     * Unlocks the current input stream if it was previously locked with
     * LockInputStream() (no-op otherwise).
     *
     */
    void UnlockInputStream();

    /**
     * Unlock output stream.
     *
     * Unlocks the current output stream if it was previously locked with
     * LockOutputStream() (no-op otherwise).
     */
    void UnlockOutputStream();

    /**
     * @brief Write output (wide version).
     *
     * Writes given string to the configured output stream.
     *
     * @param str String to write to output stream.
     */
    void Write(const std::wstring& str);
};

/**
 * @brief Mouse coordinate reporting mode.
 */
enum class MouseCoordinateMode {
    Character, ///< Coordinates are character col/row.
    Pixel ///< Coordinates are pixel x y.
};

/**
 * @brief Mouse coordinate event selection.
 */
enum class MouseEventMode {
    None, ///< Disables mouse reporting.
    ButtonOnly, ///< Only report on mouse down/up.
    All, ///< Report all mouse events.
    Highlight ///< Enable highlight mode.
};

/**
 * @brief Button event type.
 *
 * Indicates whether a button was depressed or released.
 */
enum class ButtonEventType{ None, Press, Release };

/**
 * @brief Modifier key.
 */
enum class ModKey{ Shift, Ctrl, Meta };

/**
 * @brief Holds mouse event data.
 */
struct MouseEventData {

    /**
     * @brief Mouse position.
     */
    Vector2D position{};

    /**
     * @brief Mouse button pressed/released.
     */
    std::uint8_t m_button;

    /**
     * @brief Modifier keys pressed.
     */
    std::vector<ModKey> m_modifiers;

    /**
     * @brief Mouse even type.
     */
    ButtonEventType eventType{ButtonEventType::None};
};

/**
 * @brief Input event type.
 */
enum class InputEventType { Mouse, Keyboard };

typedef std::function<void(const std::wstring& keyData)> KeyboardEventCallback;
typedef std::function<void(const MouseEventData& mouseData)> MouseEventCallback;

/**
 * @brief Input parser.
 *
 * Reads from input and triggers events for mouse and keyboard data with parsed
 * ansi sequences detailing things such as mouse position and special keycodes.
 */
class InputParser {

    KeyboardEventCallback m_onKeyEvent;
    MouseEventCallback m_onMouseEvent;

public:

    /**
     * @brief Default constructor.
     */
    InputParser() = default;

    /**
     * @brief Reads input and triggers events.
     *
     * Blocking, reads input via IO::Read and returns the result.
     *
     * Normally one character at a time is read, but may return multiple:
     *
     * If the read character looks like the start of an escape sequence, more
     * characters will be read in attempt to parse it. If parsing is successful
     * either an event will be triggered, or the entire sequence is returned.
     * If parsing is not successful, the read buffer so far is returned.
     */
    void Read();

    /**
     * @brief Set 'onkey' event.
     *
     * This triggers when Read() finishes parsing characters from input.
     *
     * @param ev Function to run on key event.
     */
    void SetOnKeyEvent(const KeyboardEventCallback& ev);

    /**
     * @brief Set 'onmouse' event.
     *
     * This triggers when Read() successfully finishes parsing mouse data.
     *
     * @param ev Function to run on mouse event.
     */
    void SetOnMouseEvent(const MouseEventCallback& ev);

};

// Functions -------------------------------------------------------------------

/**
 * @brief Clear screen.
 *
 * Erases all characters on the screen.
 */
void ClearScreen();

/**
 * @brief Get cursor position.
 *
 * Lets you determine where the cursor currently is on the screen.
 *
 * @note In order for this to function properly, Terminal::Unbuffer() must be
 *       called before calling this function, as it relies on a response from
 *       the terminal which we must be able to read without having to wait for 
 *       a buffer flush (like when the user presses enter for instance). The 
 *       Application class will already do this for you on stdin, but if you're
 *       using this directly, it's something to consider.
 *
 * @return Returns the current cursor's screen coordinates.
 */
Vector2D GetCursorPos();

/**
 * @brief Hides the cursor.
 */
void HideCursor();

/**
 * @brief Sets the cursor position.
 *
 * @note When relative is true, this function uses GetCursorPos to get the
 *       current cursor position and as such, the same note as for that
 *       function applies in terms of calling Terminal::Unbuffer().
 *
 * @param pos New cursor position.
 * @param relative (optional, default false) Moves cursor relative to current
 *                 position if true. Otherwise, absolute coordinates are used.
 */
void SetCursorPos(const Vector2D& pos, bool relative = false);

/**
 * @brief Sets background color (16 color version).
 *
 * Sets the current background color for any following output to the given 
 * color. This version uses standard 16 colors.
 *
 * @param color New output background color.
 */
void SetBackgroundColor(StandardColor color);

/**
 * @brief Sets background color (rgb color version).
 *
 * Sets the current background color for any following output to the given
 * color. This version uses red/green/blue values.
 *
 * Note that some terminals may round to the nearest color on a 256 color
 * palette, whilst others might support true color.
 *
 * @param r New output background color red value.
 * @param g New output background color green value.
 * @param b New output background color blue value.
 */
void SetBackgroundColor(std::uint8_t r, std::uint8_t g, std::uint8_t b);

/**
 * @brief Sets foreground color (16 color version).
 *
 * Sets the current foreground color for any following output to the given 
 * color. This version uses standard 16 colors.
 *
 * @param color New output foreground color.
 */
void SetForegroundColor(StandardColor color);

/**
 * @brief Sets foreground color (rgb color version).
 *
 * Sets the current foreground color for any following output to the given
 * color. This version uses red/green/blue values.
 *
 * Note that some terminals may round to the nearest color on a 256 color
 * palette, whilst others might support true color.
 *
 * @param r New output foreground color red value.
 * @param g New output foreground color green value.
 * @param b New output foreground color blue value.
 */
void SetForegroundColor(std::uint8_t r, std::uint8_t g, std::uint8_t b);

/**
 * @brief Set mouse reporting mode.
 *
 * Causes the terminal to report on mouse coordinates. We only use the SGR and
 * SGR_PIXEL formats, which should be compatible with most terminals.
 *
 * @param eventMode Sets what type of mouse events should be reported.
 * @param coordinateMode Sets how coordinates should be reported.
 */
void SetMouseReporting(
    MouseEventMode eventMode,
    MouseCoordinateMode coordinateMode = MouseCoordinateMode::Character
);

/**
 * @brief Makes cursor visible.
 */
void ShowCursor();

/**
 * Convert Color enum to ANSI background color code.
 *
 * @param c Color to convert.
 * @return Color as ansi background color SGR code.
 */
std::wstring ToAnsiBgColor(StandardColor c);

/**
 * Convert Color enum to ANSI foreground color code.
 *
 * @param c Color to convert.
 * @return Color as ansi foreground color SGR code.
 */
std::wstring ToAnsiFgColor(StandardColor c);

} // namespace neovision
