#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>

#include <cerrno>
#include <codecvt>
#include <cstring>
#include <locale>
#include <string>

#include "neovision/application.h"
#include "neovision/crt.h"

#include "neovision/term.h"

namespace neovision {

Terminal::Terminal()
{
}

Terminal::~Terminal()
{
    Cleanup();
}

void Terminal::Cleanup()
{
    if (m_initialized != true) return;
    Restore();
    if (m_fd) close(m_fd);
}

void Terminal::Initialize()
{
    // Open terminal file handle with non-blocking I/O.
    m_fd = open("/dev/tty", O_NONBLOCK|O_RDWR);
    // Register signal handlers.
    sigemptyset(&m_resizeSignal.sa_mask);
    m_resizeSignal.sa_flags = SA_RESTART;
    m_resizeSignal.sa_handler = sigWinchHandler;
    if (sigaction(SIGWINCH, &m_resizeSignal, NULL) == -1)
    {
        throw std::runtime_error(
            "Could not register terminal resize signal handler."
        );
    }
    signal(SIGINT, sigIntHandler);
    m_initialized = true;
}

void Terminal::sigIntHandler(int)
{
    // We call the actual c++ callback from here if one is registred.
    const std::shared_ptr<Application> app = TheApp.lock();
    if (app == nullptr) return; // No application instantiated yet.
    const Terminal& term = app->Term();
    if (term.m_onTerminate)
    {
        term.m_onTerminate();
    }
}

void Terminal::sigWinchHandler(int)
{
    /* This is a c-callback invoked when the SIGWINCH signal is received.
       We call the actual c++ callback from here if one is registred. */
    const std::shared_ptr<Application> app = TheApp.lock();
    if (app == nullptr) return; // No application instantiated yet.
    const Terminal& term = app->Term();
    if (term.m_onResize)
    {
        term.m_onResize();
    }
}

void Terminal::ProcessEvents()
{
    if (!m_initialized) return;
    if (!m_fd) return;
    winsize ws;
    if (ioctl(m_fd, TIOCGWINSZ, &ws) == -1)
    {
        throw std::runtime_error(
            std::string{"Error while checking for terminal resize event: "} +
            std::strerror(errno) + " (" + std::to_string(errno) + ")"
        );
    }
}

std::wstring Terminal::Read(size_t n)
{
    std::wstring result;
    if (!m_initialized) return result;
    if (!m_fd) return result;
    size_t bytesRead{0};
    std::string shortResult;
    const auto startTime = std::chrono::system_clock::now();
    while (bytesRead < n) 
    {
        const auto now = std::chrono::system_clock::now();
        const auto deltaT = std::chrono::duration_cast<std::chrono::milliseconds>(
            now - startTime
        );
        if (deltaT > m_ioTimeOut) break;
        char buffer[n + 1];
        memset(buffer, 0, n + 1);
        const size_t br = read(m_fd, buffer, n);
        if (br == -1)
        {
            if (errno == EAGAIN)
            {
                std::this_thread::sleep_for(std::chrono::microseconds{10});
            }
            else
            {
                // Should this throw instead?
                break;
            }
        }
        else if (br != 0)
        {
            bytesRead += br;
            shortResult += buffer;
        }
    }
    std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> conv;
    result = conv.from_bytes(shortResult);
    return result;
}

void Terminal::Restore()
{
    if (!m_initialized) return;
    tcsetattr(m_fd, TCSANOW, &m_savedTerminalSettings);
}

Vector2D Terminal::Size()
{
    winsize w{};
    ioctl(m_fd, TIOCGWINSZ, &w);
    return {w.ws_col, w.ws_row};
}

void Terminal::SetOnResize(TerminalEventCallback callback)
{
    m_onResize = callback;
}

void Terminal::SetOnTerminate(TerminalEventCallback callback)
{
    m_onTerminate = callback;
}

std::chrono::milliseconds Terminal::Timeout() const
{
    return m_ioTimeOut;
}

void Terminal::Timeout(const std::chrono::milliseconds& t)
{
    m_ioTimeOut = t;
}


void Terminal::Unbuffer()
{
    if (!m_initialized) return;
    if (!m_fd) return;
    termios old_tio{};
    termios new_tio{};
    tcgetattr(m_fd, &old_tio);
    new_tio = old_tio;
    new_tio.c_lflag &= (~ICANON & ~ECHO);
    tcsetattr(m_fd, TCSANOW, &new_tio);
    m_savedTerminalSettings = old_tio;
}

void Terminal::Write(const std::wstring& data)
{
    if (!m_initialized) return;
    if (!m_fd) return;
    std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> conv;
    const std::string s = conv.to_bytes(data);
    size_t bytesWritten{0};
    const auto startTime = std::chrono::system_clock::now();
    while(bytesWritten != s.size())
    {
        const auto now = std::chrono::system_clock::now();
        const auto deltaT = std::chrono::duration_cast<std::chrono::seconds>(
            now - startTime
        );
        if (deltaT > m_ioTimeOut) break;
        const size_t bw = write(m_fd, &s[0], s.size());
        if (bw == -1)
        {
            if (errno == EAGAIN)
            {
                std::this_thread::sleep_for(std::chrono::microseconds{10});
            }
            else
            {
                break; // Should this throw?
            }
        }
        bytesWritten += bw;
    }
}

} // namespace neovision
